<?php
/*
 * License: GPLv3
 * License URI: https://www.gnu.org/licenses/gpl.txt
 * Copyright 2017 Jean-Sebastien Morisset (https://surniaulula.com/)
 */

if ( ! defined( 'ABSPATH' ) ) {
	die( 'These aren\'t the droids you\'re looking for...' );
}

if ( ! class_exists( 'JSM_BFO_Filters' ) ) {

	class JSM_BFO_Filters {

		private static $instance;

		public function __construct() {
			add_action( 'init', array( __CLASS__, 'add_block_filter_hooks' ) );
		}

		/*
		 * "_check_output_buffer_*" static method callbacks are added
		 * after each filter hook to check the output buffer for a
		 * non-empty string.
		 *
		 * The "*" urlencoded text is used to extract the previous hook
		 * prority and name:
		 *
		 *	preg_match( '/^_check_output_buffer_\[([0-9]+)\](.+)$/',
		 *		urldecode( $method_name ), $matches );
		 */
		public static function __callStatic( $method_name, $args ) {	// since PHP 5.3
			if ( strpos( $method_name, '_check_output_buffer_' ) === 0 ) {
				array_unshift( $args, $method_name );
				return call_user_func_array( array( __CLASS__, '_check_output_buffer' ), $args );
			}
		}

		public static function &get_instance() {
			if ( ! isset( self::$instance ) ) {
				self::$instance = new self;
			}
			return self::$instance;
		}

		/*
		 * Hooked to the WordPress 'init' action. Loop through each
		 * filter name ("the_content" by default) and add a start
		 * filter, which starts the output buffer, adds a check filter
		 * after each hook, and adds a stop output buffer filter at the
		 * end.
		 */
		public static function add_block_filter_hooks() {
			global $wp_actions;
			foreach ( self::get_block_filter_names() as $filter_name ) {	// hooks 'the_content' by default
				if ( empty( $wp_actions[$filter_name] ) ) {		// make sure it's a filter, not an action (does not match for 'all')
					if ( ! has_filter( $filter_name, array( __CLASS__, 'start_output_buffer' ) ) ) {
						add_filter( $filter_name, array( __CLASS__, 'start_output_buffer' ), self::get_min_prio(), 1 );
					}
				}
			}
		}

		/*
		 * Runs at the beginning of a filter to start the PHP output
		 * buffer, add a check filter after each hook, and add a stop
		 * output buffer at the end. When the 'all' filter is hooked in
		 * add_block_filter_hooks(), this method will be called for
		 * actions as well, so check $wp_actions again to exclude
		 * actions.
		 */
		public static function start_output_buffer( $value ) {
			global $wp_actions;
			$filter_name = current_filter();
			if ( empty( $wp_actions[$filter_name] ) ) {			// double-check - make sure it's a filter, not an action 
				if ( ob_start() ) {
					static $do_once = array();			// only check output the first time a filter is applied
					if ( ! isset( $do_once[$filter_name] ) ) {	// add a new "_check_output_buffer_*" hook after each hook
						$do_once[$filter_name] = false;		// prevent hooks from being added again on a second filter run
						self::add_check_output_hooks( $filter_name );
					} elseif ( $do_once[$filter_name] === false ) {	// if filter is run a second time, remove the check hooks
						$do_once[$filter_name] = true;		// prevent hooks from being added or removed on third filter run
						self::remove_check_output_hooks( $filter_name );
					}
					add_filter( $filter_name, array( __CLASS__, 'stop_output_buffer' ), self::get_max_int(), 1 );
				}
			}
			return $value;
		}

		/*
		 * Called by start_output_buffer() at the beginning of a filter
		 * to add a check callback after each hook. This method is
		 * called once per filter run, so applying the same filter a
		 * second time, etc., will not re-check hook output - one check
		 * is enough to detect and report any output problems.
		 */
		public static function add_check_output_hooks( $filter_name ) {
			global $wp_filter;
			if ( isset( $wp_filter[$filter_name]->callbacks ) ) {
				foreach ( $wp_filter[$filter_name]->callbacks as $hook_prio => &$hook_group ) {
					$new_hook_group = array();						// create a new group to insert a check after each hook
					foreach ( $hook_group as $hook_ref => $hook_info ) {
						$new_hook_group[$hook_ref] = $hook_info;			// add the original callback first, followed by the check
						$hook_name = self::get_hook_function_name( $hook_info );	// create a human readable class / method name
						if ( $hook_name === '' ) {					// just in case
							continue;
						} elseif ( strpos( $hook_name, __CLASS__.'::' ) === 0 ) {	// exclude our own class / methods from being checked
							continue;
						}
						$check_hook_ref = $hook_ref.'_check_output_buffer';		// extend the previous hook ref for visual clue
						$check_hook_arg = urlencode( '['.$hook_prio.']'.$hook_name );	// pass previous hook priority and name
						$new_hook_group[$check_hook_ref] = array(
							'function' => array(
								__CLASS__,
								'_check_output_buffer_'.$check_hook_arg,	// calls the __callStatic() method
							), 
							'accepted_args' => 1,
						);
					}
					$hook_group = $new_hook_group;
				}
			}
		}

		/*
		 * Remove the output check hooks if/when a filter is applied a
		 * second time. The output filter is only checked for a
		 * non-empty string on the first filter run.
		 */
		public static function remove_check_output_hooks( $filter_name ) {
			global $wp_filter;
			foreach ( $wp_filter[$filter_name]->callbacks as $hook_prio => &$hook_group ) {
				foreach ( $hook_group as $hook_ref => $hook_info ) {
					if ( strpos( $hook_ref, '_'.__CLASS__.'_check_output_buffer' ) !== false ) {
						unset( $hook_group[$hook_ref] );
					}
				}
			}
		}

		/*
		 * Runs at the end of a filter to clean (truncate) and end
		 * (terminate) the output buffer.
		 */
		public static function stop_output_buffer( $value ) {
			ob_end_clean();
			return $value;
		}

		/*
		 * Called by the add_block_filter_hooks() method, which is
		 * hooked to the WordPress 'init' action.
		 * 
		 * Get a list of filters to hook and block / check their output
		 * (only hooks "the_content" filter by default). Define the
		 * JSM_BFO_FILTER_NAMES constant to customize the filter names.
		 * You can also use the special WordPress 'all' filter to hook
		 * and check all filters.
		 */
		public static function get_block_filter_names() {
			static $filter_names = null;
			if ( $filter_names === null ) {
				$filter_names = array();
				if ( defined( 'JSM_BFO_FILTER_NAMES' ) ) {
					if ( is_array( JSM_BFO_FILTER_NAMES ) ) {
						$filter_names = JSM_BFO_FILTER_NAMES;
					} elseif ( is_string( JSM_BFO_FILTER_NAMES ) ) {
						$filter_names = array( JSM_BFO_FILTER_NAMES );
					}
				} else {
					$filter_names = array( 'the_content' );	// just hook the content filter by default
				}
			}
			return $filter_names;
		}

		/*
		 * Get a human readable class/method/function name from the
		 * hook array. Called by the add_check_output_hooks() method,
		 * which uses the encoded hook name as part of the
		 * "_check_output_buffer_*" callback name (to pass the previous
		 * hook priority and name for the error message).
		 */
		public static function get_hook_function_name( array $hook_info ) {
			$hook_name = '';
			if ( ! isset( $hook_info['function'] ) ) {		// just in case
				return $hook_name;				// stop here - return an empty string
			} elseif ( is_array( $hook_info['function'] ) ) {	// hook is a class / method
				$class_name = '';
				$function_name = '';
				if ( is_object( $hook_info['function'][0] ) ) {
					$class_name = get_class( $hook_info['function'][0] );
				} elseif ( is_string( $hook_info['function'][0] ) ) {
					$class_name = $hook_info['function'][0];
				}
				if ( is_string( $hook_info['function'][1] ) ) {
					$function_name = $hook_info['function'][1];

				}
				return $class_name.'::'.$function_name;
			} elseif ( is_string ( $hook_info['function'] ) ) {	// hook is a function
				return $hook_info['function'];
			}
			return $hook_name;
		}
		
		public static function get_min_int() {
			return defined( 'PHP_INT_MIN' ) ? PHP_INT_MIN : -2147483648;	// available since PHP 7.0.0
		}

		public static function get_max_int() {
			return defined( 'PHP_INT_MAX' ) ? PHP_INT_MAX : 2147483647;	// available since PHP 5.0.2
		}

		/*
		 * Called by the __callStatic() method after each filter hook.
		 * Checks the output buffer for any non-empty string.
		 */
		private static function _check_output_buffer( $method_name, $value ) {
			$output = ob_get_contents();
			if ( $output !== '' ) {	// the previous hook has (incorrectly) provided some output
				ob_clean();	// clean / truncate the output buffer

				$error_text = __( 'The "%1$s" filter hook with priority %2$d in the "%3$s" filter has mistakenly provided some output.',
					'jsm-block-filter-output' ).' '.
				__( 'All WordPress filter hooks must return their data / text and not send it to the output.',
					'jsm-block-filter-output' ).' '.
				__( 'Please contact the author of that filter hook and report this as a coding error / bug.',
					'jsm-block-filter-output' );

				if ( preg_match( '/^_check_output_buffer_\[([0-9]+)\](.+)$/', urldecode( $method_name ), $matches ) ) {
					$error_msg = sprintf( $error_text, $matches[2], $matches[1], current_filter() );
					error_log( $error_msg );
					/*
					 * Filters are rarely applied on the admin / back-end side, 
					 * but if they are, then take advantage of this and show a 
					 * notice. :)
					 */
					if ( is_admin() ) {
						$lib_dir = trailingslashit( realpath( dirname( __FILE__ ) ) );
						require_once $lib_dir.'com/notice.php';	// load the SucomNotice class
						$notice = new SucomNotice( null, 'bfo', 'Block Filter Output', 'jsm-block-filter-output' );
						$notice->err( $error_msg );
					}
				}
			}
			return $value;
		}
	}

	JSM_BFO_Filters::get_instance();
}

?>
